"""imbalanced_spdf is a python package for imbalanced_spdf data handling in machine learning."""

__author__ = """Rahul Goswami"""
__email__ = 'yuvrajiro@gmail.com'
__version__ = '0.0.1'

from .ensemble import SPBaDF, SPBoDF
from .tree import tree